from typing import Any, Dict


class GeralSettings:
    __rootdir = "rootdir"
    __is_ascii = "ascii"
    __is_color = "color"
    __diffdown = "diffdown"
    __sidesize = "sidesize"
    __langdef = "lang"

    defaults = {
        __rootdir: "",
        __is_ascii: False,
        __is_color: True,
        __diffdown: True,
        __sidesize: 80,
        __langdef: ""
    }

    def __init__(self):
        self.data: Dict[str, Any] = {}
        for key in GeralSettings.defaults:
            self.data[key] = GeralSettings.defaults[key]

    def __set(self, key: str, value: Any):
        self.data[key] = value
        return self
    
    def __get(self, key: str) -> str:
        if key not in self.defaults:
            raise ValueError(f"Key {key} not found in GeralSettings")
        return self.data.get(key, GeralSettings.defaults[key])
        

    def to_dict(self) -> Dict[str, Any]:
        return self.data
        
    def from_dict(self, data: Dict[str, Any]):
        self.data = data
        return self

    def __str__(self) -> str:
        return str(self.data)


    def get_rootdir(self) -> str:
        return self.__get(self.__rootdir)
    
    def set_rootdir(self, value: str):
        self.__set(self.__rootdir, value)
        return self
    
    def get_is_ascii(self):
        return self.__get(self.__is_ascii)

    def set_is_ascii(self, value: bool):
        self.__set(self.__is_ascii, value)
        return self
    
    def get_is_colored(self):
        return self.__get(self.__is_color)
    
    def set_is_colored(self, value: bool):
        self.__set(self.__is_color, value)
        return self
    
    def get_is_diff_down(self):
        return self.__get(self.__diffdown)
    
    def set_is_diff_down(self, value: bool):
        self.__set(self.__diffdown, value)
        return self
    
    def get_side_size(self):
        return self.__get(self.__sidesize)
    
    def set_side_size(self, value: int):
        self.__set(self.__sidesize, value)
        return self
    
    def get_lang_def(self):
        return self.__get(self.__langdef)
    
    def set_lang_def(self, value: str):
        self.__set(self.__langdef, value)
        return self
    



import curses
from .play.fmt import Fmt
from .run.basic import ExecutionResult, Param, Unit
from .execution import Execution
from .run.diff import Diff
from .util.ftext import Sentence, Token
from typing import List, Optional
from .play.frame import Frame
from .play.floating import Floating

from .run.wdir import Wdir
from .run.report import Report
from .settings.settings_parser import SettingsParser
import os

class CDiff:

    def __init__(self, wdir: Wdir, param: Param.Basic):
        self.param = param
        self.results: List[Token] = []
        self.wdir = wdir
        self.exit = False
        self.index = 0

        self.init = 0   # index of first line to show
        self.length = 1  # length of diff
        self.space = 0  # dy space for draw

        self.finished = False
        self.resumes: List[str] = []
        self.first_error = -1

        self.sp = SettingsParser()
        self.settings = self.sp.load_settings()
        self.first_loop = True
        self.warning: Optional[Floating] = None

    def save_settings(self):
        self.settings.geral.set_is_diff_down(self.param.is_up_down)
        self.sp.save_settings()

    def set_exit(self):
        self.exit = True

    def end_processing(self):
        return len(self.results) == len(self.wdir.unit_list)

    def get_color(self, unit: Unit):
        if unit.result == ExecutionResult.SUCCESS:
            return "G"
        if unit.result == ExecutionResult.WRONG_OUTPUT:
            return "R"
        if unit.result == ExecutionResult.EXECUTION_ERROR:
            return "Y"
        return ""

    def sucesso(self):
        out1 = [
            r"  _____ __ __     __    ___  _____ _____  ___  ",
            r" / ___/|  |  |   /  ]  /  _]/ ___// ___/ /   \ ",
            r"(   \_ |  |  |  /  /  /  [_(   \_(   \_ |     |",
            r" \__  ||  |  | /  /  |    _]\__  |\__  ||  O  |",
            r" /  \ ||  :  |/   \_ |   [_ /  \ |/  \ ||     |",
            r" \    ||     |\     ||     |\    |\    ||     |",
            r"  \___| \__,_| \____||_____| \___| \___| \___/ ",
        ]

        out = out1
        _, cols = Fmt.get_size()
        for i, line in enumerate(out):
            Fmt.write(i + 4, 1, Sentence().addf("g", line).center(cols - 2, Token(" ", " ")))


    def draw_scrollbar(self):
        y_init = 4
        if not self.has_any_error():
            return   
        tr = "╮"
        br = "╯"
        vbar = "│"
        bar = []

        if self.length > self.space:
            total = self.space
            _begin = False
            _end = False
            if self.init == 0:
                _begin = True
            if self.init == self.length - self.space:
                _end = True

            # pre = int((self.init / self.length) * total)
            # pos = int((max(0, self.length - self.init - self.space) / self.length) * total)            
            # mid = total - pre - pos
            pre = int((self.init / self.length) * total)
            mid = int((self.space / self.length) * total)
            pos = (max(0, total - pre - mid))

            if _begin:
                pre -= 1
            if _end:
                pos -= 1

            if self.init > 0 and pre == 0:
                pre = 1
                pos -= 1

            if _begin:
                bar.append(tr)
            for _ in range(pre):
                bar.append(vbar)
            for _ in range(mid):
                bar.append("┃")
            for _ in range(pos):
                bar.append(vbar)
            if _end:
                bar.append(br)

        elif self.length < self.space:
            bar.append(tr)
            for i in range(self.length - 2):
                bar.append(vbar)
            bar.append(br)

        _, cols = Fmt.get_size()
        for i in range(len(bar)):
            Fmt.write(i + y_init, cols - 1, Sentence().add(bar[i]))


    def draw_top_line(self):
        # construir mais uma solução
        if len(self.results) < len(self.wdir.unit_list):
            index = len(self.results)
            unit = self.wdir.unit_list[index]
            if self.wdir.solver is None:
                return
            unit.result = Execution.run_unit(self.wdir.solver, unit)
            symbol = ExecutionResult.get_symbol(unit.result)
            color = self.get_color(unit)
            self.results.append(Token(symbol.text, color))
            if color != "G" and self.first_error == -1:
                self.first_error = len(self.results) - 1
                self.index = len(self.results) - 1


        _, cols = Fmt.get_size()
        frame = Frame(0, 0).set_size(3, cols)
        pwd = os.getcwd()
        intro = self.wdir.resume()
        base = os.sep.join(pwd.split(os.sep)[:-1])
        last = pwd.split(os.sep)[-1]
        frame.set_header(Sentence().add(" ").add(base).add("/").addf("y", last).add("   ").add(intro).add(" ") , "<")
        unit = self.wdir.unit_list[self.index]
        frame.set_footer(Sentence().add(" ").add(unit.str(False)), "")
        frame.draw()

        x = 0
        i = 0
        for i, symbol in enumerate(self.results):
            foco = i == self.index
            extrap = "" if not foco else ">"
            extras = "" if not foco else "<"
            label = Sentence().add(extrap).addf(symbol.fmt, str(i).zfill(2)).add(symbol).add(extras)
            frame.write(0, x + 1 - len(extrap), label)
            x += 5
        
    def draw_guide_line(self):
        cmds = (Sentence()
        .add(" ")
        .addf("/B", "ReCompile").addf("B","[r]")
        .add(" ")
        .addf("/B", "SelectCase").addf("B", "[a, d]")
        .add(" ")
        .addf("/B", "Scroll").addf("B", "[w, s]")
        .add(" ")
        .addf("/B", "DiffMode").addf("B", "[D]")
        .add(" ")
        )
        _, cols = Fmt.get_size()
        Fmt.write(3, 0, cmds.center(cols, Token(" ", "K")))

    def has_any_error(self):
        results = [unit.result for unit in self.wdir.unit_list]
        if ExecutionResult.EXECUTION_ERROR not in results and ExecutionResult.WRONG_OUTPUT not in results:
            return False
        return True          

    def draw_diff(self, unit: Unit):
        lines, cols = Fmt.get_size()
        self.space = lines - 4
        frame = Frame(3, -1).set_inner(self.space, cols - 1).set_border_square()
        # frame.draw()
        if not self.has_any_error():
            self.sucesso()
            return        
        Report.set_terminal_size(cols)
        if self.param.is_up_down:
            line_list = Diff.mount_up_down_diff(unit, curses=True)
        else:
            line_list = Diff.mount_side_by_side_diff(unit, curses=True)

        self.length = max(1, len(line_list))

        if self.length - self.init < self.space:
            self.init = max(0, self.length - self.space)

        if self.init >= self.length:
            self.init = self.length - 1

        if self.init < self.length:
            line_list = line_list[self.init:]
        for i, line in enumerate(line_list):
            frame.write(i, 0, Sentence().add(line))
        return

    def load_autoload_warning(self):
        # if not self.wdir.autoload:
        #     return
        warning = Floating().set_header(" Atenção ").warning()
        warning.put_text("Os seguintes arquivos foram carregados automaticamente")
        solver = self.wdir.solver
        solvers = [] if solver is None else solver.path_list
        warning.put_text("")
        loaded = Sentence().add("Códigos ")
        for i, file in enumerate(solvers):
            if i != 0:
                loaded.add(", ")
            loaded.addf("g", os.path.basename(file))
        warning.put_sentence(loaded)
        sources = self.wdir.source_list
        loaded = Sentence().add("Testes ")
        for i, file in enumerate(sources):
            if i != 0:
                loaded.add(", ")
            loaded.addf("y", os.path.basename(file))

        warning.put_sentence(loaded)
        warning.put_text("")
        warning.put_text("Caso queira remover algum arquivo")
        warning.put_text("renomeie sua extensão para .txt")
        warning.put_text("Ou execute o run manualmente passando os arquivos desejados")
        warning.put_text("")
        warning.put_sentence(Sentence().addf("c", "tko run <cod> <cod> cases.tio")) 
        warning.put_text("")

        self.warning = warning

    def main(self, scr):
        curses.curs_set(0)  # Esconde o cursor
        Fmt.init_colors()  # Inicializa as cores
        Fmt.set_scr(scr)  # Define o scr como global
        while not self.exit:
            if self.first_loop:
                self.first_loop = False
                self.load_autoload_warning()
            Fmt.erase()
            self.draw_top_line()
            unit = self.wdir.unit_list[self.index]
            self.draw_diff(unit)
            self.draw_scrollbar()
            self.draw_guide_line()

            if self.warning is not None and self.warning.is_enable():
                self.warning.draw()
            if not self.end_processing():
                Fmt.refresh()
                continue

            if self.warning is not None and self.warning.is_enable():
                input = self.warning.get_input()
            else:
                input = Fmt.getch()

            if input == ord('q'):
                self.set_exit()
            elif input == curses.KEY_LEFT or input == ord('a'):
                self.index = max(0, self.index - 1)
                self.init = 0
            elif input == curses.KEY_RIGHT or input == ord('d'):
                self.index = min(len(self.results) - 1, self.index + 1)
                self.init = 0
            elif input == curses.KEY_DOWN or input == ord('s'):
                self.init += 1
            elif input == curses.KEY_UP or input == ord('w'):
                self.init = max(0, self.init - 1)
            elif input == ord('D'):
                self.param.is_up_down = not self.param.is_up_down
                self.save_settings()
                self.init = 0
            elif input == ord('r'):
                if self.wdir.solver is not None:
                    self.wdir.solver.prepare_exec()
                self.results = []
                self.first_error = -1
                

    def run(self):
        curses.wrapper(self.main)



from typing import List, Optional
from .quest import Quest
from ..util.ftext import Sentence


class Cluster:
    def __init__(self, line_number: int = 0, title: str = "", key: str = "", color: Optional[str] = None):
        self.line_number = line_number
        self.title: str = title
        self.key: str = key
        self.quests: List[Quest] = []
        self.color: Optional[str] = color

    def __str__(self):
        line = str(self.line_number).rjust(3)
        quests_size = str(len(self.quests)).rjust(2, "0")
        key = "" if self.key == self.title else self.key + " "
        return f"{line} {quests_size} {key}{self.title}"
    
    def get_grade_color(self) -> str:
        perc = self.get_percent()
        if perc == 0:
            return "m"
        if perc < 50:
            return "r"
        if perc < 100:
            return "y"
        return "g"

    def get_percent(self):
        total = 0
        for q in self.quests:
            total += q.get_percent()
        return total // len(self.quests)

    def get_resume_by_percent(self) -> Sentence:
        return Sentence().addf(self.get_grade_color(), f"{self.get_percent()}%".rjust(4))

    def get_resume_by_quests(self):
        total = len(self.quests)
        count = len([q for q in self.quests if q.is_complete()])
        return Sentence().addf(self.get_grade_color(), f"({count}/{total})")
        



from typing import List, Optional
import os
import shutil
import subprocess

from .run.wdir import Wdir
from .run.basic import DiffMode, ExecutionResult, CompilerError, Param
from .run.diff import Diff
from .util.ftext import Sentence, Token

from .run.report import Report
from .util.term_color import term_print
from .util.symbols import symbols

from .run.writer import Writer
from .util.runner import Runner
from .cdiff import CDiff
from .execution import Execution



class FilterMode:

    @staticmethod
    def deep_copy_and_change_dir():
        # path to ~/.tko_filter
        filter_path = os.path.join(os.path.expanduser("~"), ".tko_filter")

        # verify if filter command is available
        if shutil.which("filter") is None:
            print("ERROR: filter command not found")
            print("Install feno with 'pip install feno'")
            exit(1)

        subprocess.run(["filter", "-rf", ".", "-o", filter_path])

        os.chdir(filter_path)


class Run:

    def __init__(self, target_list: List[str], exec_cmd: Optional[str], param: Param.Basic):
        self.target_list: List[str] = target_list
        self.exec_cmd: Optional[str] = exec_cmd
        self.param: Param.Basic = param
        self.wdir: Wdir = Wdir()
        self.curses: bool = False

    def set_curses(self, value:bool=True):
        self.curses = value
        return self

    def execute(self):
        self.__remove_duplicates()
        self.__change_targets_to_filter_mode()
        if not self.__build_wdir():
            return
        if self.__missing_target():
            return
        if self.__list_mode():
            return
        if self.__free_run():
            return
        self.__diff_mode()
        return

    def __remove_duplicates(self):
        # remove duplicates in target list keeping the order
        self.target_list = list(dict.fromkeys(self.target_list))

    def __change_targets_to_filter_mode(self):
        if self.param.filter:
            old_dir = os.getcwd()

            term_print(Report.centralize(" Entering in filter mode ", "═"))
            FilterMode.deep_copy_and_change_dir()  
            # search for target outside . dir and redirect target
            new_target_list = []
            for target in self.target_list:
                if ".." in target:
                    new_target_list.append(os.path.normpath(os.path.join(old_dir, target)))
                elif os.path.exists(target):
                    new_target_list.append(target)
            self.target_list = new_target_list

    def __print_top_line(self):
        if self.wdir is None:
            return

        term_print(symbols.opening + self.wdir.resume(), end="")
        term_print(" [", end="")
        first = True
        for unit in self.wdir.unit_list:
            if first:
                first = False
            else:
                term_print(" ", end="")
            solver = self.wdir.solver
            if solver is None:
                raise ValueError("Solver empty")
            unit.result = Execution.run_unit(solver, unit)
            term_print(Sentence() + ExecutionResult.get_symbol(unit.result), end="")
        term_print("]")

    def __print_diff(self):
        if self.wdir is None:
            return
        
        if self.param.diff_mode == DiffMode.QUIET:
            return
        
        results = [unit.result for unit in self.wdir.unit_list]
        if ExecutionResult.EXECUTION_ERROR not in results and ExecutionResult.WRONG_OUTPUT not in results:
            return
        
        if not self.param.compact:
            for elem in self.wdir.unit_list_resume():
                term_print(elem)
        
        if self.param.diff_mode == DiffMode.FIRST:
            # printing only the first wrong case
            wrong = [unit for unit in self.wdir.unit_list if unit.result != ExecutionResult.SUCCESS][0]
            if self.param.is_up_down:
                for line in Diff.mount_up_down_diff(wrong):
                    term_print(line)
            else:
                for line in Diff.mount_side_by_side_diff(wrong):
                    term_print(line)
            return

        if self.param.diff_mode == DiffMode.ALL:
            for unit in self.wdir.unit_list:
                if unit.result != ExecutionResult.SUCCESS:
                    if self.param.is_up_down:
                        for line in Diff.mount_up_down_diff(unit):
                            term_print(line)
                    else:
                        for line in Diff.mount_side_by_side_diff(unit):
                            term_print(line)

    def __build_wdir(self) -> bool:
        try:
            self.wdir = Wdir().set_curses(self.curses).set_target_list(self.target_list).set_cmd(self.exec_cmd).build().filter(self.param)
        except CompilerError as e:
            print(e)
            return False
        except FileNotFoundError as e:
            print(e)
            return False
        return True

    def __missing_target(self) -> bool:
        if self.wdir is None:
            return False
        # no solver and no test cases
        if self.wdir.solver is None and len(self.wdir.unit_list) == 0:
            term_print(Sentence().addf("", "fail: ") + "No solver or tests found.")
            return True
        return False
    
    def __list_mode(self) -> bool:
        if self.wdir is None:
            return False

        # list mode
        if self.wdir.solver is None and len(self.wdir.unit_list) > 0:
            term_print(Report.centralize(" No solvers found. Listing Test Cases ", Token("╌")), flush=True)
            term_print(self.wdir.resume())
            for line in self.wdir.unit_list_resume():
                term_print(line)
            return True
        return False

    def __free_run(self) -> bool:
        if self.wdir is None:
            return False
        # free run mode
        if self.wdir.solver is not None and len(self.wdir.unit_list) == 0:
            t = Report.centralize(Sentence() + " No test cases found. Running: " + self.wdir.solver.executable + " ", symbols.hbar)
            term_print(t, flush=True)
            # force print to terminal
            Runner.free_run(self.wdir.solver.executable)
            return True
        return False

    def __diff_mode(self):
        if self.wdir is None:
            return
        
        if self.curses:
            cdiff = CDiff(self.wdir, self.param)
            cdiff.run()
            return
        term_print(Report.centralize(" Running solver against test cases ", "═"))
        self.__print_top_line()
        self.__print_diff()


class Build:

    def __init__(self, target_out: str, source_list: List[str], param: Param.Manip, to_force: bool):
        self.target_out = target_out
        self.source_list = source_list
        self.param = param
        self.to_force = to_force

    def execute(self):
        try:
            wdir = Wdir().set_sources(self.source_list).build()
            wdir.manipulate(self.param)
            Writer.save_target(self.target_out, wdir.unit_list, self.to_force)
        except FileNotFoundError as e:
            print(str(e))
            return False
        return True



from typing import List, Dict, Any
from ..util.ftext import Sentence

class Flag:
    def __init__(self):
        self._name: str = ""
        self._text: str = ""  # description
        self._char: str = ""
        self._values: List[str] = ["0", "1"]
        self._index: int = 0
        self._location: str = ""
        self._bool = True  # many options

    def name(self, _name):
        self._name = _name
        return self

    def text(self, _text):
        self._text = _text
        return self

    def char(self, _key):
        self._char = _key
        return self

    def values(self, _values: List[str]):
        self._values = _values
        return self

    def index(self, _index):
        self._index = _index
        return self

    def location(self, value: str):
        self._location = value
        return self

    def toggle(self):
        self._index = (self._index + 1) % len(self._values)
        return self

    def many(self):
        self._bool = False
        return self
    
    def bool(self):
        self._bool = True
        return self
    
    def is_bool(self):
        return self._bool

    def get_location(self) -> str:
        return self._location

    def get_value(self) -> str:
        return self._values[self._index]

    def is_true(self):
        return self.get_value() == "1"

    def get_name(self) -> str:
        return self._name

    def get_description(self) -> str:
        return self._text

    def get_char(self) -> str:
        return self._char

    def get_index(self) -> int:
        return self._index

    def get_color(self) -> str:
        if self.get_value() == "1":
            return Flags.flag_on.get_value()
        return Flags.flag_off.get_value()

    def get_toggle_sentence(self, pad: int = 0) -> Sentence:
        if not self._bool:
            name = Sentence().addf(self.get_value(), f"{self._name}".ljust(pad))
            value = Sentence().add(f"[{self.get_char()}]").add(name).add(f"{self.get_value()}".rjust(2))
            return value
            
        char = self.get_char()
        text = self.get_name()
        color = self.get_color()
        extra = Sentence()
        if pad > 0:
            extra.addf(color, (pad - len(text)) * " ")
        visual = "━─"
        if self.is_true():
            visual = "─━"
        value = Sentence().add(f"[{char}]").addf(color, f"{visual} ").addf(color + "/", text).add(extra)
        return value
    

class Flags:
    count = Flag().name("Count").char("C").values(["1", "0"]).text("Mostra a contagem de tarefas").location("left")
    down = Flag().name("Down").char("D").values(["1", "0"]).text("Mostra o caminho para a tarefa baixadas").location("left")
    minimum = Flag().name("Minimum").char("M").text("Mostra a nota mínima para a tarefa").location("left")
    opt = Flag().name("Opt").char("O").text("Mostra tarefas opcionais").location("left")
    relative = Flag().name("PathRel").char("P").values(["0", "1"]).text("Mostra o path relativo para os arquivos").location("left")
    xp = Flag().name("Exp").char("E").text("Mostra a xp obtida").location("left")
    percent = Flag().name("Percent").char("%").text("Mostra a porcentagem de tarefas").location("left")
    dots = Flag().name("Dots").char(".").values(["1", "0"]).text("Mostra o preenchimento com pontos").location("left")
    group_prog = Flag().name("Group").values(["1", "0"]).text("Mostra a barra de progresso dos grupos")
    quest_prog = Flag().name("Quest").values(["1", "0"]).text("Mostra a barra de progresso das missões")
    admin = Flag().name("Admin").char("A").text("Mostra todas as missões e grupos").location("left")

    flags_bar = Flag().name("FlagsBar").char("F").values(["0", "1"]).text("Mostra a barra de flags").location("top")
    help_bar = Flag().name("HelpBar").char("H").values(["1", "0"]).text("Mostra a barra de ajuda").location("top")
    skills_bar = Flag().name("SkillsBar").char("S").values(["0", "1"]).text("Mostra a barra de skills").location("top")

    focus     = Flag().name("Selected").char("z").values(["B", "R", "G", "Y", "wK", "kW"]).text("Cor do item em foco").many().location("left")
    prog_done = Flag().name("ProgDone").char("x").values(["g", "b", "c", "k", "w"]).text("Progresso Done").many().location("left")
    prog_todo = Flag().name("ProgTodo").char("c").values(["y", "m", "r", "k", "w"]).text("Progresso Todo").many().location("left")
    flag_on   = Flag().name("Toggle_1").char("v").values(["G", "W", "B", "C", "wK", "kW"]).text("Flag True").many().location("left")
    flag_off  = Flag().name("Toggle_0").char("b").values(["Y", "R", "M", "wK", "kW"]).text("Flag False").many().location("left")
    cmds      = Flag().name("Cmds").char("n").values(["B", "C", "M", "Y", "wK", "kW"]).text("CMDS").many()
    skill_done = Flag().name("ExpeDone").char("y").values(["G", "B", "C", "wK", "kW"]).text("Skill Done").many().location("left")
    skill_todo = Flag().name("ExpeTodo").char("u").values(["Y", "R", "M", "wK", "kW"]).text("Skill Todo").many().location("left")
    main_done = Flag().name("MainDone").char("i").values(["B", "G", "C", "wK", "kW"]).text("Main Done").many().location("left")
    main_todo = Flag().name("MainTodo").char("p").values(["M", "R", "Y", "wK", "kW"]).text("Main Todo").many().location("left")


class FlagsMan:
    def __init__(self, data: Dict[str, int]):
        self.flags: Dict[str, Flag] = {}
        self.left: List[Flag] = []
        self.top: List[Flag] = []

        for flag in Flags.__dict__.values():
            if isinstance(flag, Flag):
                self.flags[flag.get_name()] = flag
                if flag.get_location() == "left":
                    self.left.append(flag)
                elif flag.get_location() == "top":
                    self.top.append(flag)

        for key, _index in data.items():
            if key in self.flags:
                self.flags[key].index(_index)

    def get_data(self) -> Dict[str, int]:
        data = {}
        for name, flag in self.flags.items():
            data[name] = flag.get_index()
        return data




import json
from typing import Any, Dict
from .rep_settings import RepSettings
from .geral_settings import GeralSettings

class Settings:
    def __init__(self):
        self.reps: Dict[str, RepSettings] = {}
        self.geral = GeralSettings()
        self.reps["fup"] = RepSettings().load_defaults().set_url("https://github.com/qxcodefup/arcade/blob/master/Readme.md")
        self.reps["ed"] = RepSettings().load_defaults().set_url("https://github.com/qxcodeed/arcade/blob/master/Readme.md")
        self.reps["poo"] = RepSettings().load_defaults().set_url("https://github.com/qxcodepoo/arcade/blob/master/Readme.md")

    def get_repo(self, course: str) -> RepSettings:
        if course not in self.reps:
            raise ValueError(f"Course {course} not found in settings")
        return self.reps[course]
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "reps": {k: v.to_dict() for k, v in self.reps.items()},
            "geral": self.geral.to_dict()
        }

    def from_dict(self, data: Dict[str, Any]):
        self.reps = {k: RepSettings().from_dict(v) for k, v in data.get("reps", {}).items()}
        self.geral = GeralSettings().from_dict(data.get("geral", {}))
        return self
    
    def save_to_json(self, file: str):
        with open(file, "w", encoding="utf-8") as f:
            json.dump(self.to_dict(), f, indent=4)

    def __str__(self):
        output = ["Repositories:"]
        maxlen = max([len(key) for key in self.reps])
        for key in self.reps:
            prefix = f"- {key.ljust(maxlen)}"
            if self.reps[key].file and self.reps[key].url:
                output.append(f"{prefix} : dual   : {self.reps[key].url} ; {self.reps[key].file}")
            elif self.reps[key].url:
                output.append(f"{prefix} : remote : {self.reps[key].url}")
            else:
                output.append(f"{prefix} : local  : {self.reps[key].file}")
        return "\n".join(output)

